#include "windows.h"
#include "ntgui.h"
#include "ntdialog.h"

SHIELD3_ICON ICON shield3.ico
SHIELD4_ICON ICON shield4.ico
SHIELD2_ICON ICON shield2.ico
SHIELD1_ICON ICON shield1.ico
LAMBDA_ICON ICON lambda.ico
LAMBDA2_ICON ICON lambda2.ico
EDWIN_ICON ICON edwin.ico
LIAR1_ICON ICON liar1.ico
LIAR2_ICON ICON liar2.ico
LIAR3_ICON ICON liar3.ico
GRAPHICS_ICON ICON graphics.ico
CONSES_ICON ICON conses.ico
ENVIRONMENT_ICON ICON envir1.ico
COFFEE_ICON ICON coffee.ico
MINCER_ICON ICON mincer.ico
BCH_ICO ICON bch.ico

GC_CURSOR  CURSOR gc.cur

1 VERSIONINFO
	FILEVERSION	0,0,0,65
	PRODUCTVERSION	7,8,0,65
	FILEOS		VOS__WINDOWS32
	FILETYPE	VFT_APP
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "040904E4"
    BEGIN
	VALUE "CompanyName", "GNU Project"
	VALUE "FileDescription", "MIT/GNU Scheme Microcode"
	VALUE "FileVersion", "14.11"
	VALUE "InternalName", "SCHEME"
	VALUE "LegalCopyright", "Copyright Massachusetts Institute of Technology 1993-2003"
	VALUE "OriginalFilename", "SCHEME.EXE"
	VALUE "ProductName", "MIT/GNU Scheme"
	VALUE "ProductVersion", "7.8.0"
    END
  END

  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x0409, 1251
  END
END

#ifdef __WATCOMC__
#include "ntdialog.dlg"
#else
RCINCLUDE NTDIALOG.DLG
#endif
